@@ -55,42 +55,37 @@
 #include <private/qv4compiler_p.h>
 #include <private/qv4compileddata_p.h>
 #include <private/qqmljsmemorypool_p.h>
+#include <private/qqmljsfixedpoolarray_p.h>
 #include <private/qv4codegen_p.h>
 #include <private/qv4compiler_p.h>
-#include <private/qqmljslexer_p.h>
 #include <QTextStream>
 #include <QCoreApplication>
 
-#ifndef V4_BOOTSTRAP
-#include <private/qqmlpropertycache_p.h>
-#endif
-
 QT_BEGIN_NAMESPACE
 
 class QQmlPropertyCache;
 class QQmlContextData;
 class QQmlTypeNameCache;
+struct QQmlIRLoader;
 
 namespace QmlIR {
 
 struct Document;
-struct IRLoader;
 
 template <typename T>
 struct PoolList
 {
     PoolList()
-        : first(0)
-        , last(0)
-        , count(0)
+        : first(nullptr)
+        , last(nullptr)
     {}
 
     T *first;
     T *last;
-    int count;
+    int count = 0;
 
     int append(T *item) {
-        item->next = 0;
+        item->next = nullptr;
         if (last)
             last->next = item;
         else
@@ -110,7 +105,7 @@ struct PoolList
     template <typename Sortable, typename Base, Sortable Base::*sortMember>
     T *findSortedInsertionPoint(T *item) const
     {
-        T *insertPos = 0;
+        T *insertPos = nullptr;
 
         for (T *it = first; it; it = it->next) {
             if (!(it->*sortMember <= item->*sortMember))
@@ -163,6 +158,13 @@ struct PoolList
     }
 
     struct Iterator {
+        // turn Iterator into a proper iterator
+        using iterator_category = std::forward_iterator_tag;
+        using value_type = T;
+        using difference_type = ptrdiff_t;
+        using pointer = T *;
+        using reference = T &;
+
         T *ptr;
 
         explicit Iterator(T *p) : ptr(p) {}
@@ -183,8 +185,15 @@ struct PoolList
             return *ptr;
         }
 
-        void operator++() {
+        Iterator& operator++() {
             ptr = ptr->next;
+            return *this;
+        }
+
+        Iterator operator++(int) {
+            Iterator that {ptr};
+            ptr = ptr->next;
+            return that;
         }
 
         bool operator==(const Iterator &rhs) const {
@@ -198,89 +207,55 @@ struct PoolList
 
     Iterator begin() { return Iterator(first); }
     Iterator end() { return Iterator(nullptr); }
-};
 
-template <typename T>
-class FixedPoolArray
-{
-    T *data;
-public:
-    int count;
-
-    FixedPoolArray()
-        : data(0)
-        , count(0)
-    {}
-
-    void allocate(QQmlJS::MemoryPool *pool, int size)
-    {
-        count = size;
-        data = reinterpret_cast<T*>(pool->allocate(count * sizeof(T)));
-    }
-
-    void allocate(QQmlJS::MemoryPool *pool, const QVector<T> &vector)
-    {
-        count = vector.count();
-        data = reinterpret_cast<T*>(pool->allocate(count * sizeof(T)));
-
-        if (QTypeInfo<T>::isComplex) {
-            for (int i = 0; i < count; ++i)
-                new (data + i) T(vector.at(i));
-        } else {
-            memcpy(data, static_cast<const void*>(vector.constData()), count * sizeof(T));
-        }
-    }
-
-    template <typename Container>
-    void allocate(QQmlJS::MemoryPool *pool, const Container &container)
-    {
-        count = container.count();
-        data = reinterpret_cast<T*>(pool->allocate(count * sizeof(T)));
-        typename Container::ConstIterator it = container.constBegin();
-        for (int i = 0; i < count; ++i)
-            new (data + i) T(*it++);
-    }
-
-    const T &at(int index) const {
-        Q_ASSERT(index >= 0 && index < count);
-        return data[index];
-    }
-
-    T &operator[](int index) {
-        Q_ASSERT(index >= 0 && index < count);
-        return data[index];
-    }
-
-
-    int indexOf(const T &value) const {
-        for (int i = 0; i < count; ++i)
-            if (data[i] == value)
-                return i;
-        return -1;
-    }
-
-    const T *begin() const { return data; }
-    const T *end() const { return data + count; }
+    using iterator = Iterator;
 };
 
 struct Object;
 
-struct SignalParameter : public QV4::CompiledData::Parameter
+struct EnumValue : public QV4::CompiledData::EnumValue
 {
-    SignalParameter *next;
+    EnumValue *next;
+};
+
+struct Enum
+{
+    int nameIndex;
+    QV4::CompiledData::Location location;
+    PoolList<EnumValue> *enumValues;
+
+    int enumValueCount() const { return enumValues->count; }
+    PoolList<EnumValue>::Iterator enumValuesBegin() const { return enumValues->begin(); }
+    PoolList<EnumValue>::Iterator enumValuesEnd() const { return enumValues->end(); }
+
+    Enum *next;
+};
+
+
+struct Parameter : public QV4::CompiledData::Parameter
+{
+    Parameter *next;
+
+    bool init(QV4::Compiler::JSUnitGenerator *stringGenerator, const QString &parameterName, const QString &typeName);
+    static bool init(QV4::CompiledData::Parameter *param, const QV4::Compiler::JSUnitGenerator *stringGenerator,
+                     int parameterNameIndex, int typeNameIndex);
+    static bool initType(QV4::CompiledData::ParameterType *paramType,
+                         const QV4::Compiler::JSUnitGenerator *stringGenerator, int typeNameIndex);
+
+    static QV4::CompiledData::BuiltinType stringToBuiltinType(const QString &typeName);
 };
 
 struct Signal
 {
     int nameIndex;
     QV4::CompiledData::Location location;
-    PoolList<SignalParameter> *parameters;
+    PoolList<Parameter> *parameters;
 
     QStringList parameterStringList(const QV4::Compiler::StringTableGenerator *stringPool) const;
 
     int parameterCount() const { return parameters->count; }
-    PoolList<SignalParameter>::Iterator parametersBegin() const { return parameters->begin(); }
-    PoolList<SignalParameter>::Iterator parametersEnd() const { return parameters->end(); }
+    PoolList<Parameter>::Iterator parametersBegin() const { return parameters->begin(); }
+    PoolList<Parameter>::Iterator parametersEnd() const { return parameters->end(); }
 
     Signal *next;
 };
@@ -300,48 +275,49 @@ struct Binding : public QV4::CompiledData::Binding
     Binding *next;
 };
 
+struct InlineComponent : public QV4::CompiledData::InlineComponent
+{
+    InlineComponent *next;
+};
+
 struct Alias : public QV4::CompiledData::Alias
 {
     Alias *next;
 };
 
+struct RequiredPropertyExtraData : public QV4::CompiledData::RequiredPropertyExtraData
+{
+    RequiredPropertyExtraData *next;
+};
+
 struct Function
 {
-    QQmlJS::AST::FunctionDeclaration *functionDeclaration;
     QV4::CompiledData::Location location;
     int nameIndex;
     quint32 index; // index in parsedQML::functions
-    FixedPoolArray<int> formals;
+    QQmlJS::FixedPoolArray<Parameter> formals;
+    QV4::CompiledData::ParameterType returnType;
 
     // --- QQmlPropertyCacheCreator interface
-    const int *formalsBegin() const { return formals.begin(); }
-    const int *formalsEnd() const { return formals.end(); }
+    const Parameter *formalsBegin() const { return formals.begin(); }
+    const Parameter *formalsEnd() const { return formals.end(); }
     // ---
 
     Function *next;
 };
 
-struct Q_QML_PRIVATE_EXPORT CompiledFunctionOrExpression
+struct Q_QMLCOMPILER_PRIVATE_EXPORT CompiledFunctionOrExpression
 {
     CompiledFunctionOrExpression()
-        : node(0)
-        , nameIndex(0)
-        , disableAcceleratedLookups(false)
-        , next(0)
     {}
-    CompiledFunctionOrExpression(QQmlJS::AST::Node *n)
-        : node(n)
-        , nameIndex(0)
-        , disableAcceleratedLookups(false)
-        , next(0)
-    {}
-    QQmlJS::AST::Node *node; // FunctionDeclaration, Statement or Expression
-    quint32 nameIndex;
-    bool disableAcceleratedLookups;
-    CompiledFunctionOrExpression *next;
+
+    QQmlJS::AST::Node *parentNode = nullptr; // FunctionDeclaration, Statement or Expression
+    QQmlJS::AST::Node *node = nullptr; // FunctionDeclaration, Statement or Expression
+    quint32 nameIndex = 0;
+    CompiledFunctionOrExpression *next = nullptr;
 };
 
-struct Q_QML_PRIVATE_EXPORT Object
+struct Q_QMLCOMPILER_PRIVATE_EXPORT Object
 {
     Q_DECLARE_TR_FUNCTIONS(Object)
 public:
@@ -350,6 +326,7 @@ public:
     int id;
     int indexOfDefaultPropertyOrAlias;
     bool defaultPropertyIsAlias;
+    bool isInlineComponent = false;
     quint32 flags;
 
     QV4::CompiledData::Location location;
@@ -359,12 +336,19 @@ public:
     int propertyCount() const { return properties->count; }
     Alias *firstAlias() const { return aliases->first; }
     int aliasCount() const { return aliases->count; }
+    const Enum *firstEnum() const { return qmlEnums->first; }
+    int enumCount() const { return qmlEnums->count; }
     const Signal *firstSignal() const { return qmlSignals->first; }
     int signalCount() const { return qmlSignals->count; }
     Binding *firstBinding() const { return bindings->first; }
     int bindingCount() const { return bindings->count; }
     const Function *firstFunction() const { return functions->first; }
     int functionCount() const { return functions->count; }
+    const InlineComponent *inlineComponent() const { return inlineComponents->first; }
+    int inlineComponentCount() const { return inlineComponents->count; }
+    const RequiredPropertyExtraData *requiredPropertyExtraData() const {return requiredPropertyExtraDatas->first; }
+    int requiredPropertyExtraDataCount() const { return requiredPropertyExtraDatas->count; }
+    void simplifyRequiredProperties();
 
     PoolList<Binding>::Iterator bindingsBegin() const { return bindings->begin(); }
     PoolList<Binding>::Iterator bindingsEnd() const { return bindings->end(); }
@@ -372,23 +356,30 @@ public:
     PoolList<Property>::Iterator propertiesEnd() const { return properties->end(); }
     PoolList<Alias>::Iterator aliasesBegin() const { return aliases->begin(); }
     PoolList<Alias>::Iterator aliasesEnd() const { return aliases->end(); }
+    PoolList<Enum>::Iterator enumsBegin() const { return qmlEnums->begin(); }
+    PoolList<Enum>::Iterator enumsEnd() const { return qmlEnums->end(); }
     PoolList<Signal>::Iterator signalsBegin() const { return qmlSignals->begin(); }
     PoolList<Signal>::Iterator signalsEnd() const { return qmlSignals->end(); }
     PoolList<Function>::Iterator functionsBegin() const { return functions->begin(); }
     PoolList<Function>::Iterator functionsEnd() const { return functions->end(); }
+    PoolList<InlineComponent>::Iterator inlineComponentsBegin() const { return inlineComponents->begin(); }
+    PoolList<InlineComponent>::Iterator inlineComponentsEnd() const { return inlineComponents->end(); }
+    PoolList<RequiredPropertyExtraData>::Iterator requiredPropertyExtraDataBegin() const {return requiredPropertyExtraDatas->begin(); }
+    PoolList<RequiredPropertyExtraData>::Iterator requiredPropertyExtraDataEnd() const {return requiredPropertyExtraDatas->end(); }
 
     // If set, then declarations for this object (and init bindings for these) should go into the
     // specified object. Used for declarations inside group properties.
     Object *declarationsOverride;
 
-    void init(QQmlJS::MemoryPool *pool, int typeNameIndex, int idIndex, const QQmlJS::AST::SourceLocation &location = QQmlJS::AST::SourceLocation());
-
-    QString sanityCheckFunctionNames(const QSet<QString> &illegalNames, QQmlJS::AST::SourceLocation *errorLocation);
+    void init(QQmlJS::MemoryPool *pool, int typeNameIndex, int idIndex, const QQmlJS::SourceLocation &location = QQmlJS::SourceLocation());
 
+    QString appendEnum(Enum *enumeration);
     QString appendSignal(Signal *signal);
-    QString appendProperty(Property *prop, const QString &propertyName, bool isDefaultProperty, const QQmlJS::AST::SourceLocation &defaultToken, QQmlJS::AST::SourceLocation *errorLocation);
-    QString appendAlias(Alias *prop, const QString &aliasName, bool isDefaultProperty, const QQmlJS::AST::SourceLocation &defaultToken, QQmlJS::AST::SourceLocation *errorLocation);
+    QString appendProperty(Property *prop, const QString &propertyName, bool isDefaultProperty, const QQmlJS::SourceLocation &defaultToken, QQmlJS::SourceLocation *errorLocation);
+    QString appendAlias(Alias *prop, const QString &aliasName, bool isDefaultProperty, const QQmlJS::SourceLocation &defaultToken, QQmlJS::SourceLocation *errorLocation);
     void appendFunction(QmlIR::Function *f);
+    void appendInlineComponent(InlineComponent *ic);
+    void appendRequiredPropertyExtraData(RequiredPropertyExtraData *extraData);
 
     QString appendBinding(Binding *b, bool isListBinding);
     Binding *findBinding(quint32 nameIndex) const;
@@ -397,23 +388,26 @@ public:
     QString bindingAsString(Document *doc, int scriptIndex) const;
 
     PoolList<CompiledFunctionOrExpression> *functionsAndExpressions;
-    FixedPoolArray<int> runtimeFunctionIndices;
+    QQmlJS::FixedPoolArray<int> runtimeFunctionIndices;
 
-    FixedPoolArray<quint32> namedObjectsInComponent;
-    int namedObjectsInComponentCount() const { return namedObjectsInComponent.count; }
+    QQmlJS::FixedPoolArray<quint32> namedObjectsInComponent;
+    int namedObjectsInComponentCount() const { return namedObjectsInComponent.size(); }
     const quint32 *namedObjectsInComponentTable() const { return namedObjectsInComponent.begin(); }
 
 private:
-    friend struct IRLoader;
+    friend struct ::QQmlIRLoader;
 
     PoolList<Property> *properties;
     PoolList<Alias> *aliases;
+    PoolList<Enum> *qmlEnums;
     PoolList<Signal> *qmlSignals;
     PoolList<Binding> *bindings;
     PoolList<Function> *functions;
+    PoolList<InlineComponent> *inlineComponents;
+    PoolList<RequiredPropertyExtraData> *requiredPropertyExtraDatas;
 };
 
-struct Q_QML_PRIVATE_EXPORT Pragma
+struct Q_QMLCOMPILER_PRIVATE_EXPORT Pragma
 {
     enum PragmaType {
         PragmaSingleton = 0x1
@@ -423,42 +417,42 @@ struct Q_QML_PRIVATE_EXPORT Pragma
     QV4::CompiledData::Location location;
 };
 
-struct Q_QML_PRIVATE_EXPORT Document
+struct Q_QMLCOMPILER_PRIVATE_EXPORT Document
 {
     Document(bool debugMode);
     QString code;
     QQmlJS::Engine jsParserEngine;
-    QV4::IR::Module jsModule;
+    QV4::Compiler::Module jsModule;
     QList<const QV4::CompiledData::Import *> imports;
     QList<Pragma*> pragmas;
     QQmlJS::AST::UiProgram *program;
-    int indexOfRootObject;
     QVector<Object*> objects;
     QV4::Compiler::JSUnitGenerator jsGenerator;
 
-    QQmlRefPointer<QV4::CompiledData::CompilationUnit> javaScriptCompilationUnit;
+    QV4::CompiledData::CompilationUnit javaScriptCompilationUnit;
 
     int registerString(const QString &str) { return jsGenerator.registerString(str); }
     QString stringAt(int index) const { return jsGenerator.stringForIndex(index); }
 
-    static void removeScriptPragmas(QString &script);
+    int objectCount() const {return objects.size();}
+    Object* objectAt(int i) const {return objects.at(i);}
 };
 
-struct Q_QML_PRIVATE_EXPORT ScriptDirectivesCollector : public QQmlJS::Directives
+class Q_QMLCOMPILER_PRIVATE_EXPORT ScriptDirectivesCollector : public QQmlJS::Directives
 {
-    ScriptDirectivesCollector(QQmlJS::Engine *engine, QV4::Compiler::JSUnitGenerator *unitGenerator);
-
+    QmlIR::Document *document;
     QQmlJS::Engine *engine;
     QV4::Compiler::JSUnitGenerator *jsGenerator;
-    QList<const QV4::CompiledData::Import *> imports;
-    bool hasPragmaLibrary;
+
+public:
+    ScriptDirectivesCollector(QmlIR::Document *doc);
 
     void pragmaLibrary() override;
     void importFile(const QString &jsfile, const QString &module, int lineNumber, int column) override;
     void importModule(const QString &uri, const QString &version, const QString &module, int lineNumber, int column) override;
 };
 
-struct Q_QML_PRIVATE_EXPORT IRBuilder : public QQmlJS::AST::Visitor
+struct Q_QMLCOMPILER_PRIVATE_EXPORT IRBuilder : public QQmlJS::AST::Visitor
 {
     Q_DECLARE_TR_FUNCTIONS(QQmlCodeGenerator)
 public:
@@ -482,43 +476,59 @@ public:
     bool visit(QQmlJS::AST::UiArrayBinding *ast) override;
     bool visit(QQmlJS::AST::UiObjectBinding *ast) override;
     bool visit(QQmlJS::AST::UiObjectDefinition *ast) override;
+    bool visit(QQmlJS::AST::UiInlineComponent *ast) override;
+    bool visit(QQmlJS::AST::UiEnumDeclaration *ast) override;
     bool visit(QQmlJS::AST::UiPublicMember *ast) override;
     bool visit(QQmlJS::AST::UiScriptBinding *ast) override;
     bool visit(QQmlJS::AST::UiSourceElement *ast) override;
+    bool visit(QQmlJS::AST::UiRequired *ast) override;
+
+    void throwRecursionDepthError() override
+    {
+        recordError(QQmlJS::SourceLocation(),
+                    QStringLiteral("Maximum statement or expression depth exceeded"));
+    }
 
     void accept(QQmlJS::AST::Node *node);
 
     // returns index in _objects
-    bool defineQMLObject(int *objectIndex, QQmlJS::AST::UiQualifiedId *qualifiedTypeNameId, const QQmlJS::AST::SourceLocation &location, QQmlJS::AST::UiObjectInitializer *initializer, Object *declarationsOverride = 0);
-    bool defineQMLObject(int *objectIndex, QQmlJS::AST::UiObjectDefinition *node, Object *declarationsOverride = 0)
+    bool defineQMLObject(int *objectIndex, QQmlJS::AST::UiQualifiedId *qualifiedTypeNameId, const QQmlJS::SourceLocation &location, QQmlJS::AST::UiObjectInitializer *initializer, Object *declarationsOverride = nullptr);
+    bool defineQMLObject(int *objectIndex, QQmlJS::AST::UiObjectDefinition *node, Object *declarationsOverride = nullptr)
     { return defineQMLObject(objectIndex, node->qualifiedTypeNameId, node->qualifiedTypeNameId->firstSourceLocation(), node->initializer, declarationsOverride); }
 
     static QString asString(QQmlJS::AST::UiQualifiedId *node);
     QStringRef asStringRef(QQmlJS::AST::Node *node);
-    static void extractVersion(QStringRef string, int *maj, int *min);
-    QStringRef textRefAt(const QQmlJS::AST::SourceLocation &loc) const
+    static void extractVersion(const QStringRef &string, int *maj, int *min);
+    QStringRef textRefAt(const QQmlJS::SourceLocation &loc) const
     { return QStringRef(&sourceCode, loc.offset, loc.length); }
-    QStringRef textRefAt(const QQmlJS::AST::SourceLocation &first,
-                         const QQmlJS::AST::SourceLocation &last) const;
+    QStringRef textRefAt(const QQmlJS::SourceLocation &first,
+                         const QQmlJS::SourceLocation &last) const;
 
-    void setBindingValue(QV4::CompiledData::Binding *binding, QQmlJS::AST::Statement *statement);
+    void setBindingValue(QV4::CompiledData::Binding *binding, QQmlJS::AST::Statement *statement,
+                         QQmlJS::AST::Node *parentNode);
+    void tryGeneratingTranslationBinding(const QStringRef &base, QQmlJS::AST::ArgumentList *args, QV4::CompiledData::Binding *binding);
 
-    void appendBinding(QQmlJS::AST::UiQualifiedId *name, QQmlJS::AST::Statement *value);
+    void appendBinding(QQmlJS::AST::UiQualifiedId *name, QQmlJS::AST::Statement *value,
+                       QQmlJS::AST::Node *parentNode);
     void appendBinding(QQmlJS::AST::UiQualifiedId *name, int objectIndex, bool isOnAssignment = false);
-    void appendBinding(const QQmlJS::AST::SourceLocation &qualifiedNameLocation, const QQmlJS::AST::SourceLocation &nameLocation, quint32 propertyNameIndex, QQmlJS::AST::Statement *value);
-    void appendBinding(const QQmlJS::AST::SourceLocation &qualifiedNameLocation, const QQmlJS::AST::SourceLocation &nameLocation, quint32 propertyNameIndex, int objectIndex, bool isListItem = false, bool isOnAssignment = false);
+    void appendBinding(const QQmlJS::SourceLocation &qualifiedNameLocation,
+                       const QQmlJS::SourceLocation &nameLocation, quint32 propertyNameIndex,
+                       QQmlJS::AST::Statement *value, QQmlJS::AST::Node *parentNode);
+    void appendBinding(const QQmlJS::SourceLocation &qualifiedNameLocation,
+                       const QQmlJS::SourceLocation &nameLocation, quint32 propertyNameIndex,
+                       int objectIndex, bool isListItem = false, bool isOnAssignment = false);
 
     bool appendAlias(QQmlJS::AST::UiPublicMember *node);
 
     Object *bindingsTarget() const;
 
-    bool setId(const QQmlJS::AST::SourceLocation &idLocation, QQmlJS::AST::Statement *value);
+    bool setId(const QQmlJS::SourceLocation &idLocation, QQmlJS::AST::Statement *value);
 
     // resolves qualified name (font.pixelSize for example) and returns the last name along
     // with the object any right-hand-side of a binding should apply to.
     bool resolveQualifiedId(QQmlJS::AST::UiQualifiedId **nameToResolve, Object **object, bool onAssignment = false);
 
-    void recordError(const QQmlJS::AST::SourceLocation &location, const QString &description);
+    void recordError(const QQmlJS::SourceLocation &location, const QString &description);
 
     quint32 registerString(const QString &str) const { return jsGenerator->registerString(str); }
     template <typename _Tp> _Tp *New() { return pool->New<_Tp>(); }
@@ -528,9 +538,12 @@ public:
     static bool isStatementNodeScript(QQmlJS::AST::Statement *statement);
     static bool isRedundantNullInitializerForPropertyDeclaration(Property *property, QQmlJS::AST::Statement *statement);
 
+    QString sanityCheckFunctionNames(Object *obj, const QSet<QString> &illegalNames, QQmlJS::SourceLocation *errorLocation);
+
     QList<QQmlJS::DiagnosticMessage> errors;
 
     QSet<QString> illegalNames;
+    QSet<QString> inlineComponentsNames;
 
     QList<const QV4::CompiledData::Import *> _imports;
     QList<Pragma*> _pragmas;
@@ -544,97 +557,32 @@ public:
     QQmlJS::MemoryPool *pool;
     QString sourceCode;
     QV4::Compiler::JSUnitGenerator *jsGenerator;
+
+    bool insideInlineComponent = false;
 };
 
-struct Q_QML_PRIVATE_EXPORT QmlUnitGenerator
+struct Q_QMLCOMPILER_PRIVATE_EXPORT QmlUnitGenerator
 {
-    QV4::CompiledData::Unit *generate(Document &output, const QV4::CompiledData::DependentTypesHasher &dependencyHasher = QV4::CompiledData::DependentTypesHasher());
+    void generate(Document &output, const QV4::CompiledData::DependentTypesHasher &dependencyHasher = QV4::CompiledData::DependentTypesHasher());
 
 private:
     typedef bool (Binding::*BindingFilter)() const;
     char *writeBindings(char *bindingPtr, const Object *o, BindingFilter filter) const;
 };
 
-#ifndef V4_BOOTSTRAP
-struct Q_QML_EXPORT PropertyResolver
+struct Q_QMLCOMPILER_PRIVATE_EXPORT JSCodeGen : public QV4::Compiler::Codegen
 {
-    PropertyResolver(const QQmlPropertyCache *cache)
-        : cache(cache)
-    {}
-
-    QQmlPropertyData *property(int index) const
-    {
-        return cache->property(index);
-    }
-
-    enum RevisionCheck {
-        CheckRevision,
-        IgnoreRevision
-    };
-
-    QQmlPropertyData *property(const QString &name, bool *notInRevision = 0, RevisionCheck check = CheckRevision) const;
-
-    // This code must match the semantics of QQmlPropertyPrivate::findSignalByName
-    QQmlPropertyData *signal(const QString &name, bool *notInRevision) const;
-
-    const QQmlPropertyCache *cache;
-};
-#endif
-
-struct Q_QML_PRIVATE_EXPORT JSCodeGen : public QQmlJS::Codegen
-{
-    JSCodeGen(const QString &fileName, const QString &sourceCode, QV4::IR::Module *jsModule,
-              QQmlJS::Engine *jsEngine, QQmlJS::AST::UiProgram *qmlRoot, QQmlTypeNameCache *imports,
-              const QV4::Compiler::StringTableGenerator *stringPool);
-
-    struct IdMapping
-    {
-        QString name;
-        int idIndex;
-        QQmlPropertyCache *type;
-    };
-    typedef QVector<IdMapping> ObjectIdMapping;
-
-    void beginContextScope(const ObjectIdMapping &objectIds, QQmlPropertyCache *contextObject);
-    void beginObjectScope(QQmlPropertyCache *scopeObject);
+    JSCodeGen(Document *document, const QSet<QString> &globalNames);
 
     // Returns mapping from input functions to index in IR::Module::functions / compiledData->runtimeFunctions
     QVector<int> generateJSCodeForFunctionsAndBindings(const QList<CompiledFunctionOrExpression> &functions);
 
-protected:
-    void beginFunctionBodyHook() override;
-    QV4::IR::Expr *fallbackNameLookup(const QString &name, int line, int col) override;
+    bool generateCodeForComponents(const QVector<quint32> &componentRoots);
+    bool compileComponent(int contextObject);
+    bool compileJavaScriptCodeInObjectsRecursively(int objectIndex, int scopeObjectIndex);
 
 private:
-    QQmlPropertyData *lookupQmlCompliantProperty(QQmlPropertyCache *cache, const QString &name, bool *propertyExistsButForceNameLookup = 0);
-
-    QString sourceCode;
-    QQmlJS::Engine *jsEngine; // needed for memory pool
-    QQmlJS::AST::UiProgram *qmlRoot;
-    QQmlTypeNameCache *imports;
-    const QV4::Compiler::StringTableGenerator *stringPool;
-
-    bool _disableAcceleratedLookups;
-    ObjectIdMapping _idObjects;
-    QQmlPropertyCache *_contextObject;
-    QQmlPropertyCache *_scopeObject;
-    int _qmlContextTemp;
-    int _importedScriptsTemp;
-};
-
-struct Q_QML_PRIVATE_EXPORT IRLoader {
-    IRLoader(const QV4::CompiledData::Unit *unit, QmlIR::Document *output);
-
-    void load();
-
-private:
-    QmlIR::Object *loadObject(const QV4::CompiledData::Object *serializedObject);
-
-    template <typename _Tp> _Tp *New() { return pool->New<_Tp>(); }
-
-    const QV4::CompiledData::Unit *unit;
-    QmlIR::Document *output;
-    QQmlJS::MemoryPool *pool;
+    Document *document;
 };
 
 } // namespace QmlIR
